/*
 * Copyright (C) 2013-2020 Argonne National Laboratory, Department of Energy,
 *                    UChicago Argonne, LLC and The HDF Group.
 * All rights reserved.
 *
 * The full copyright notice, including terms governing use, modification,
 * and redistribution, is contained in the COPYING file that can be
 * found at the root of the source code distribution tree.
 */

#ifndef MERCURY_LOG_H
#define MERCURY_LOG_H

#include "mercury_util_config.h"

#include <stdio.h>

/*************************************/
/* Public Type and Struct Definition */
/*************************************/

/* Available log types, additional log types should be added to that list by
 * order of verbosity. Format is:
 * - enum type
 * - type name
 */
#define HG_LOG_TYPES                                                           \
    X(HG_LOG_TYPE_NONE, "")           /*!< no log */                           \
    X(HG_LOG_TYPE_ERROR, "error")     /*!< error log type */                   \
    X(HG_LOG_TYPE_WARNING, "warning") /*!< warning log type */                 \
    X(HG_LOG_TYPE_DEBUG, "debug")     /*!< debug log type */                   \
    X(HG_LOG_TYPE_MAX, "")

#define X(a, b) a,
enum hg_log_type { HG_LOG_TYPES };
#undef X

/*****************/
/* Public Macros */
/*****************/

/* For compatibility */
#if defined(__STDC_VERSION__) && (__STDC_VERSION__ < 199901L)
#    if defined(__GNUC__) && (__GNUC__ >= 2)
#        define __func__ __FUNCTION__
#    else
#        define __func__ "<unknown>"
#    endif
#elif defined(_WIN32)
#    define __func__ __FUNCTION__
#endif

/* Log macro */
#define HG_LOG_WRITE(log_mask, log_type, module_name, ...)                     \
    do {                                                                       \
        if (log_mask >= log_type)                                              \
            hg_log_write(log_type, module_name, __FILE__, __LINE__, __func__,  \
                __VA_ARGS__);                                                  \
    } while (0)

/*********************/
/* Public Prototypes */
/*********************/

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Get the log type from the log name string.
 *
 * \param log_name [IN]         null terminated string
 *
 * \return log type enum value
 */
HG_UTIL_PUBLIC enum hg_log_type
hg_log_name_to_type(const char *log_name);

/**
 * Set the logging function.
 *
 * \param log_func [IN]         pointer to function
 */
HG_UTIL_PUBLIC void
hg_log_set_func(int (*log_func)(FILE *stream, const char *format, ...));

/**
 * Set the stream for error output.
 *
 * \param stream [IN/OUT]       pointer to stream
 */
HG_UTIL_PUBLIC void
hg_log_set_stream_error(FILE *stream);

/**
 * Set the stream for warning output.
 *
 * \param stream [IN/OUT]       pointer to stream
 */
HG_UTIL_PUBLIC void
hg_log_set_stream_warning(FILE *stream);

/**
 * Set the stream for debug output.
 *
 * \param stream [IN/OUT]       pointer to stream
 */
HG_UTIL_PUBLIC void
hg_log_set_stream_debug(FILE *stream);

/**
 * Write log.
 *
 * \param log_type [IN]         log type (HG_LOG_TYPE_DEBUG, etc)
 * \param module [IN]           module name
 * \param file [IN]             file name
 * \param line [IN]             line number
 * \param func [IN]             function name
 * \param format [IN]           string format
 */
HG_UTIL_PUBLIC void
hg_log_write(enum hg_log_type log_type, const char *module, const char *file,
    unsigned int line, const char *func, const char *format, ...);

#ifdef __cplusplus
}
#endif

#endif /* MERCURY_LOG_H */
